# Copyright (c) 2013, Tom Hendrikx
# All rights reserved.
#
# See LICENSE for the license.

import atexit
import errno
import logging
from logging.handlers import SysLogHandler
import os
import resource
import signal
import sys

logger = logging.getLogger(__name__)


def daemonize(pidfile=None):
    """
    Turn the running process into a proper daemon according to PEP3143.

    Args:
    pidfile --The pidfile to create.

    """

    # Prevent core dumps
    resource.setrlimit(resource.RLIMIT_CORE, (0, 0))

    # Change working directory
    os.chdir("/")

    # Change file creation mask
    os.umask(0)

    # Detach process context: do double fork
    pid = os.fork()
    if pid > 0:
        os._exit(0)
    os.setsid()
    pid = os.fork()
    if pid > 0:
        os._exit(0)

    # Create signal handler for SIGTERM
    def terminate(signal, stack_frame):
        msg = 'Terminating on signal {}'.format(signal)
        logger.info(msg)
        raise SystemExit(msg)
    signal.signal(signal.SIGTERM, terminate)

    # Redirect input/output streams
    streams = [sys.stdin, sys.stdout, sys.stderr]
    for stream in streams:
        devnull = os.open(os.devnull, os.O_RDWR)
        os.dup2(devnull, stream.fileno())

    # Close file descriptors
    for fd in [stream.fileno() for stream in streams]:
        try:
            os.close(fd)
        except OSError, err:
            if err.errno == errno.EBADF:
                # File descriptor was not open
                pass

    # Create pidfile
    if pidfile is not None:
        pid = os.getpid()
        with open(pidfile, 'w') as f:
            f.write('{}\n'.format(pid))
            f.close()

        def remove_pid_file():
            os.remove(pidfile)

        atexit.register(remove_pid_file)

    logger.debug('Process daemonized')


def config_str2dict(option_value):
    """
    Parse the value of a config option and convert it to a dictionary.

    The configuration allows lines formatted like:
    foo = Bar:1,Baz,Flub:0.75
    This gets converted to a dictionary:
    foo = { 'Bar': 1, 'Baz': 0, 'Flub': 0.75 }

    Args:
    option_value -- The config string to parse.

    """
    dict = {}
    for key in option_value.split(','):
        if ':' in key:
            key, value = pair.split(':')
            value = float(value)
        else:
            value = 0
        dict[key] = value
    return dict


def log_to_syslog():
    """
    Configure logging to syslog.

    """
    # Get root logger
    rl = logging.getLogger()
    rl.setLevel('INFO')

    # Stderr gets critical messages (mostly config/setup issues)
    #   only when not daemonized
    stderr = logging.StreamHandler(stream=sys.stderr)
    stderr.setLevel(logging.CRITICAL)
    stderr.setFormatter(logging.Formatter(
        '%(asctime)s %(name)s: %(levelname)s %(message)s'))
    rl.addHandler(stderr)

    # All interesting data goes to syslog, using root logger's loglevel
    syslog = SysLogHandler(address='/dev/log', facility=SysLogHandler.LOG_MAIL)
    syslog.setFormatter(logging.Formatter(
        '%(name)s[%(process)d]: %(levelname)s %(message)s'))
    rl.addHandler(syslog)
    #logger.info('Logging configured')



import os.path

import pytest
from flexmock import flexmock

from client import *


def test_init():
    c = DspamClient()
    assert(c.socket == 'inet:24@localhost')
    assert(c.dlmtp_ident is None)
    assert(c.dlmtp_pass is None)


def test_init_with_args(tmpdir):
    sock = 'unix:' + os.path.join(str(tmpdir), 'dspam.sock')
    dlmtp_ident = 'bar'
    dlmtp_pass = 'baz'
    c = DspamClient(sock, dlmtp_ident, dlmtp_pass)
    assert c.socket == sock
    assert c.dlmtp_ident == dlmtp_ident
    assert c.dlmtp_pass == dlmtp_pass


@pytest.mark.parametrize('input,expected', [
    ('foo\r\n', 'foo\r\n'),
    ('foo\n', 'foo\r\n'),
    ('foo', 'foo\r\n'),
    ('foo\r', 'foo\r\r\n'),
])
def test_send(input, expected):
    sock = flexmock()
    sock.should_receive('send').once().with_args(expected)
    c = DspamClient()
    c._socket = sock
    c._send(input)


def test_read():
    sock = flexmock()
    sock.should_receive('recv').and_return(
        '', 'f', 'o', 'o', '\r', '\n', 'b', 'a', 'r', '\n', 'qux').one_by_one()
    c = DspamClient()
    c._socket = sock
    assert c._read() == ''
    assert c._read() == 'foo'
    assert c._read() == 'bar'
    assert c._peek() == 'qux'


def test_connect(monkeypatch):
    def mp_socket_connect(self, address):
        # succeed without creating a connection
        pass
    monkeypatch.setattr(socket.socket, 'connect', mp_socket_connect)

    c = DspamClient()
    flexmock(c).should_receive('_read').once().and_return(
        '220 DSPAM DLMTP 3.10.2 Authentication Required')
    c.connect()
    assert isinstance(c._socket, socket.socket)


def test_connect_unix_failed(tmpdir):
    sock = 'unix:' + os.path.join(str(tmpdir), 'dspam.sock')
    c = DspamClient(sock)
    with pytest.raises(DspamClientError):
        c.connect()


def test_connect_invalid_socketspec():
    sock = 'foo'
    c = DspamClient(sock)
    with pytest.raises(DspamClientError):
        c.connect()

    sock = 'foo:bar'
    c = DspamClient(sock)
    with pytest.raises(DspamClientError):
        c.connect()


def test_lhlo():
    c = DspamClient(dlmtp_ident='foo')
    flexmock(c).should_receive('_send').once().with_args('LHLO foo\r\n')
    (flexmock(c)
        .should_receive('_read')
        .times(6)
        .and_return('250-localhost.localdomain')
        .and_return('250-PIPELINING')
        .and_return('250-ENHANCEDSTATUSCODES')
        .and_return('250-DSPAMPROCESSMODE')
        .and_return('250-8BITMIME')
        .and_return('250 SIZE'))
    c.lhlo()
    assert c.dlmtp is True


def test_lhlo_no_dlmtp():
    c = DspamClient()
    flexmock(c).should_receive('_send').once().with_args(re.compile('^LHLO '))
    (flexmock(c)
        .should_receive('_read')
        .times(5)
        .and_return('250-localhost.localdomain')
        .and_return('250-PIPELINING')
        .and_return('250-ENHANCEDSTATUSCODES')
        .and_return('250-8BITMIME')
        .and_return('250 SIZE'))
    c.lhlo()
    assert c.dlmtp is False


def test_lhlo_unexpected_response():
    c = DspamClient()
    flexmock(c).should_receive('_send').once().with_args(re.compile('^LHLO '))
    flexmock(c).should_receive('_read').once().and_return('foo')
    with pytest.raises(DspamClientError):
        c.lhlo()


def test_mailfrom_invalid_args():
    c = DspamClient()
    with pytest.raises(DspamClientError):
        c.mailfrom(sender='foo', client_args='bar')

    c = DspamClient()
    c.dlmtp = False
    with pytest.raises(DspamClientError):
        c.mailfrom(client_args='foo')


@pytest.mark.parametrize('sender,dlmtp_ident,dlmtp_pass,expected', [
    (None, None, None, 'MAIL FROM:<>\r\n'),
    ('qux', None, None, 'MAIL FROM:<qux>\r\n'),
    ('qux', 'foo', None, 'MAIL FROM:<qux>\r\n'),
    # sender overrules dlmtp args
    ('qux', 'foo', 'bar', 'MAIL FROM:<qux>\r\n'),
    (None, 'foo', None, 'MAIL FROM:<>\r\n'),
    # no sender, dlmtp args are ok -> use them
    (None, 'foo', 'bar', 'MAIL FROM:<bar@foo>\r\n'),
    (None, None, 'bar', 'MAIL FROM:<>\r\n'),
])
def test_mailfrom_sender(sender, dlmtp_ident, dlmtp_pass, expected):
    c = DspamClient(dlmtp_ident=dlmtp_ident, dlmtp_pass=dlmtp_pass)
    flexmock(c).should_receive('_send').once().with_args(expected)
    flexmock(c).should_receive('_read').once().and_return('250 OK')
    c.mailfrom(sender)


def test_mailfrom_client_args():
    c = DspamClient()
    c.dlmtp = True
    flexmock(c).should_receive('_send').once().with_args(
        'MAIL FROM:<> DSPAMPROCESSMODE="bar"\r\n')
    flexmock(c).should_receive('_read').once().and_return('250 OK')
    c.mailfrom(client_args='bar')


def test_mailfrom_error_response():
    c = DspamClient()
    flexmock(c).should_receive('_send').once()
    flexmock(c).should_receive('_read').once().and_return(
        '451 Some error ocurred')
    with pytest.raises(DspamClientError):
        c.mailfrom()


def test_rcptto():
    c = DspamClient()
    flexmock(c).should_receive('_send').once().with_args('RCPT TO:<foo>\r\n')
    flexmock(c).should_receive('_read').once().and_return('250 2.1.5 OK')
    c.rcptto(('foo',))


def test_rcptto_multiple():
    c = DspamClient()
    flexmock(c).should_receive('_send').times(3).with_args(re.compile(
        '^RCPT TO:<\w+>\\r\\n'))
    flexmock(c).should_receive('_read').times(3).and_return('250 2.1.5 OK')
    c.rcptto(('foo', 'bar', 'qux'))


def test_rcptto_unexpected_response():
    c = DspamClient()
    flexmock(c).should_receive('_send').once().with_args(
        re.compile('^RCPT TO:<\w+>\\r\\n'))
    flexmock(c).should_receive('_read').once().and_return(
        '550 5.5.1 No such user')
    with pytest.raises(DspamClientError):
        c.rcptto(('foo'))


def test_data_unexpected_response_at_data():
    c = DspamClient()
    flexmock(c).should_receive('_send').once().with_args('DATA\r\n')
    flexmock(c).should_receive('_read').once().and_return(
        '451 Some error ocurred')
    with pytest.raises(DspamClientError):
        c.data('Sample message')


def test_data_dotstuffing():
    c = DspamClient()
    c._recipients = ['foo']
    flexmock(c).should_receive('_send').with_args('DATA\r\n')
    flexmock(c).should_receive('_read').and_return(
        '354 Enter mail, end with "." on a line by itself').and_return(
        '250 2.5.0 <foo> Message accepted for delivery')
    flexmock(c).should_receive('_send').with_args('..')
    flexmock(c).should_receive('_send').with_args('.\r\n')
    flexmock(c).should_receive('_peek').once().and_return(
        '250 2.5.0 <foo> Message ')
    c.data('.')



# Copyright (c) 2012, Tom Hendrikx
# All rights reserved.
#
# See LICENSE for the license.

from distutils.core import setup, Command

class PyTest(Command):
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        import sys,subprocess
        errno = subprocess.call(['py.test', '--verbose', '--pep8', 'dspam/'])
        raise SystemExit(errno)

setup(name='dspam-milter',
    description='Milter implementation for DSPAM',
    author='Tom Hendrikx',
    author_email='dspam-milter@whyscream.net',
    url='https://github.com/whyscream/dspam-milter',
    version='0.3',
    packages=['dspam'],
    scripts=['bin/dspam-milter'],
    data_files=[('/etc/', ['bin/dspam-milter.cfg'])],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python',
        'Topic :: Communications :: Email :: Filters'
    ],
    license='New (3-clause) BSD License',
    cmdclass={
        'test': PyTest
    }
)



# Copyright (c) 2012, Tom Hendrikx
# All rights reserved.
#
# See LICENSE for the license.

import socket
import logging
import re


class DspamClientError(Exception):
    pass

logger = logging.getLogger(__name__)


class DspamClient(object):
    """
    A DSPAM client can be used to interact with a DSPAM server.

    The client is able to speak to a DSPAM server over both a TCP or UNIX
    domain socket exposed by a running DSPAM server, and interact with it
    through its supported protocols: LMTP and DLMTP. The latter is an
    enhanced version of LMTP to facilitate some options that are not possible
    when using strict LMTP.

    Some common DSPAM operations are included in this class, custom
    operations can be built by creating a new LMTP dialog with the
    low-level LMTP commands.

    DSPAM server setup
    ==================
    To use the client to speak with a DSPAM server, the server must be
    configured to expose a TCP (dspam.conf: ServerHost, ServerPort) or
    UNIX domain socket (dspam.conf: ServerDomainSocketPath).
    The server can support mulitple modes (dspam.conf: ServerMode) for
    interaction with connecting clients. Which mode you need, depends on
    the operations you need to perform. Most of the time you'll want to
    use DLMTP though, which means that you'll also need to setup
    authentication (dspam.conf: ServerPass.<ident>).

    Python DspamClient setup
    ========================
    Each DspamClient instance needs to talk to a DSPAM server.
    You need to specify the socket where DSPAM is listening when creating
    a new instance. If you need to use DLMTP features (probably most of the
    time), you also need to pass the ident and password.

    """

    # Default configuration
    socket = 'inet:24@localhost'
    dlmtp_ident = None
    dlmtp_pass = None

    def __init__(self, socket=None, dlmtp_ident=None, dlmtp_pass=None):
        """
        Initialize new DSPAM client.

        The socket specifies where DSPAM is listening. Specify it in the form:
        unix:PATH or inet:PORT[@HOST]. For example, the default UNIX domain
        socket in dspam.conf would look like: unix:/var/run/dspam/dspam.sock,
        and the default TCP socket: inet:24@localhost.

        Args:
        socket      -- The socket on which DSPAM is listening.
        dlmtp_ident -- The authentication identifier.
        dlmtp_pass  -- The authentication password.

        """
        if socket is not None:
            self.socket = socket
        if dlmtp_ident is not None:
            self.dlmtp_ident = dlmtp_ident
        if dlmtp_pass is not None:
            self.dlmtp_pass = dlmtp_pass

        self.dlmtp = False
        self.results = {}
        # Some internal structures
        self._socket = None
        self._recipients = []

    def __del__(self):
        """
        Destroy the DSPAM client object.

        """
        if self._socket:
            self.quit()

    def _send(self, line):
        """
        Write a line of data to the server.

        Args:
        line -- A single line of data to write to the socket.

        """
        if not line.endswith('\r\n'):
            if line.endswith('\n'):
                logger.debug('Fixing bare LF before sending data to socket')
                line = line[0:-1] + '\r\n'
            else:
                logger.debug(
                    'Fixing missing CRLF before sending data to socket')
                line = line + '\r\n'
        logger.debug('Client sent: ' + line.rstrip())
        self._socket.send(line)

    def _read(self):
        """
        Read a single response line from the server.

        """
        line = ''
        finished = False
        while not finished:
            char = self._socket.recv(1)
            if char == '':
                return ''
            elif char == '\r':
                continue
            elif char == '\n':
                finished = True
                continue
            else:
                line = line + char
        logger.debug('Server sent: ' + line.rstrip())
        return line

    def _peek(self, chars=1):
        """
        Peek at the data in the server response.

        Peeking should only be done when the response can be predicted.
        Make sure that the socket will not block by requesting too
        much data from it while peeking.

        Args:
        chars -- the number of characters to peek.

        """
        line = self._socket.recv(chars, socket.MSG_PEEK)
        logger.debug('Server sent (peek): ' + line.rstrip())
        return line

    def connect(self):
        """
        Connect to TCP or domain socket, and process the server LMTP greeting.

        """
        # extract proto from socket setting
        try:
            (proto, spec) = self.socket.split(':')
        except ValueError:
            raise DspamClientError(
                'Failed to parse DSPAM socket specification, '
                'no proto found: ' + self.socket)

        if proto == 'unix':
            # connect to UNIX domain socket
            try:
                self._socket = socket.socket(
                    socket.AF_UNIX, socket.SOCK_STREAM)
                self._socket.connect(spec)
            except socket.error, err:
                self._socket = None
                raise DspamClientError(
                    'Failed to connect to DSPAM server '
                    'at socket {}: {}'.format(spec, err))
            logger.debug('Connected to DSPAM server at socket {}'.format(spec))

        elif proto == 'inet' or proto == 'inet6':
            # connect to TCP socket
            try:
                (port, host) = spec.split('@')
                port = int(port)
                if host == '':
                    host = 'localhost'
            except ValueError:
                port = int(spec)
                host = 'localhost'

            try:
                self._socket = socket.socket(
                    socket.AF_INET, socket.SOCK_STREAM)
                self._socket.connect((host, port))
            except socket.error, err:
                self._socket = None
                raise DspamClientError(
                    'Failed to connect to DSPAM server at host {} '
                    'port {}: {}'.format(host, port, err))
            logger.debug(
                'Connected to DSPAM server at host {}, port {}'.format(
                    host, port))
        else:
            raise DspamClientError(
                'Failed to parse DSPAM socket specification, '
                'unknown proto ' + proto)

        resp = self._read()
        if not resp.startswith('220'):
            raise DspamClientError(
                'Unexpected server response at connect: ' + resp)

    def lhlo(self):
        """
        Send LMTP LHLO greeting, and process the server response.

        A regular LMTP greeting is sent, and if accepted by the server, the
        capabilities it returns are parsed.

        DLMTP authentication starts here by announcing the dlmtp_ident in
        the LHLO as our hostname. When the ident is accepted and DLMTP
        mode is enabled (dspam.conf: ServerMode=dspam|auto), the
        DSPAMPROCESSMODE capability is announced by the server.
        When this capability is detected, the <DspamClient>.dlmtp flag
        will be enabled.

        """
        if self.dlmtp_ident is not None:
            host = self.dlmtp_ident
        else:
            host = socket.getfqdn()
        self._send('LHLO ' + host + '\r\n')

        finished = False
        while not finished:
            resp = self._read()
            if not resp.startswith('250'):
                raise DspamClientError(
                    'Unexpected server response at LHLO: ' + resp)
            if resp[4:20] == 'DSPAMPROCESSMODE':
                self.dlmtp = True
                logger.debug('Detected DLMTP extension in LHLO response')
            if resp[3] == ' ':
                # difference between "250-8BITMIME" and "250 SIZE"
                finished = True

    def mailfrom(self, sender=None, client_args=None):
        """
        Send LMTP MAIL FROM command, and process the server response.

        In DLMTP mode, the server expects the client to identify itself.
        Because the envelope sender is of no importance to DSPAM, the client
        is expected to send an identity and a password (dspam.conf:
        ServerPass.<ident>="<password>") in stead of the actual sender.

        When you need want DSPAM to deliver the message itself and need to
        pass the server an actual envelope sender for that, add the
        --mail-from parameter in client_args.

        When the server is setup in LMTP mode only (dspam.conf:
        ServerMode=standard), the envelope sender is a regular envelope
        sender, and is re-used when delivering the message after processing.

        Client args
        ===========
        When in DLMTP mode (and with proper auth credentials), the server
        accepts parameters specified by the client. These are in the form
        as they are passed to the command-line 'dspam' program.
        See man dspam(1) for details, and the process() or classify() methods
        in this class for simple examples.

        Args:
        sender      -- The envelope sender to use in LMTP mode.
        client_args -- DSPAM parameters to pass to the server in DLMTP mode.

        """
        if sender and client_args:
            raise DspamClientError('Arguments are mutually exclusive')

        if client_args and not self.dlmtp:
            raise DspamClientError(
                'Cannot send client args, server does not support DLMTP')

        command = 'MAIL FROM:'
        if not sender:
            if self.dlmtp_ident and self.dlmtp_pass:
                sender = self.dlmtp_pass + '@' + self.dlmtp_ident
            else:
                sender = ''
        command = command + '<' + sender + '>'

        if client_args:
            command = command + ' DSPAMPROCESSMODE="{}"'.format(client_args)

        self._send(command + '\r\n')
        resp = self._read()
        if not resp.startswith('250'):
            raise DspamClientError(
                'Unexpected server response at MAIL FROM: ' + resp)

    def rcptto(self, recipients):
        """
        Send LMTP RCPT TO command, and process the server response.

        The DSPAM server expects to find one or more valid DSPAM users as
        envelope recipients. The set recipient will be the user DSPAM
        processes mail for.

        When you need want DSPAM to deliver the message itself, and need to
        pass the server an envelope recipient for this that differs from the
        DSPAM user account name, use the --rcpt-to parameter in client_args
        at mailfrom().

        args:
        recipients -- A list of recipients

        """
        for rcpt in recipients:
            self._send('RCPT TO:<{}>\r\n'.format(rcpt))
            resp = self._read()
            if not resp.startswith('250'):
                raise DspamClientError(
                    'Unexpected server response at RCPT TO for '
                    'recipient {}: {}'.format(rcpt, resp))
            self._recipients.append(rcpt)

    def data(self, message):
        """
        Send LMTP DATA command and process the server response.

        The server response is stored as a list of dicts in
        <DspamClient>.results, keyed on the recipient name(s). Depending
        on the server return data, different formats are available:
        * LMTP mode    -- Dict containing 'accepted', a bool indicating
                          that the message was handed to the server.
        * Summary mode -- Dict containing 'username', 'result',
                          'classification', 'probability', 'confidence'
                          and 'signature'.
        * Stdout mode  -- Dict containing 'result' and 'message', the
                          complete message payload including added headers.

        The return data is always parsed and stored, independent of its format.
        If you requested a regular LMTP response, but the server
        responded with an DLMTP summary, the summary is still stored in
        <DspamClient>.results, and you will need to check the result format
        yourself and decide whether that was acceptable for your use case.
        This is due to the fact that it's possible to configure the server to
        return non LMTP responses, even when in LMTP mode (see dspam.conf:
        ServerParameters).

        Note: while processing response data in stdout mode, it's not possible
        to relate the returned messages to a specific recipient, when multiple
        recipients were specified in rcptto(). There is no guarantee
        that the message stored in <DspamClient>.results['foo'] actually
        belongs to the recipient 'foo'. If this relationship needs to be
        guaranteed, send each message with a single recipient in rcptto().

        args:
        message -- The full message payload to pass to the server.

        """
        self._send('DATA\r\n')
        resp = self._read()
        if not resp.startswith('354'):
            raise DspamClientError(
                'Unexpected server response at DATA: ' + resp)

        # Send message payload
        for line in message.split('\n'):
            if line == '.':
                # Dot stuffing
                line = '..'
            self._send(line)

        # Send end-of-data
        self._send('.\r\n')

        # Depending on server configuration, several responses are possible:
        # * Standard LMTP response code, once for each recipient:
        #   250 2.6.0 <bar> Message accepted for delivery
        # * Summary response (--deliver=summary), once for each recipient:
        #   X-DSPAM-Result: bar; result="Spam"; class="Spam"; \
        #     probability=1.0000; confidence=0.85; \
        #     signature=50c50c0f315636261418125
        #   (after the last summary line, a single dot is sent)
        # * Stdout response (--delivery=stdout), once for each recipient:
        #   X-Daemon-Classification: INNOCENT
        #   <complete mail body>
        #
        # Note that when an unknown recipient is passed in, DSPAM will simply
        #   deliver the message (dspam.conf: (Un)TrustedDeliveryAgent,
        #   DeliveryHost) unaltered and unfiltered. The response for unknown
        #   recipients will still be something indicating 'accepted'.

        peek = self._peek(24)
        if peek.startswith('250'):
            # Response is LTMP formatted
            regex = re.compile('250 \d\.\d\.\d <([^>]+)>')
            finished = False
            while not finished:
                resp = self._read()
                match = regex.match(resp)
                if not match:
                    raise DspamClientError(
                        'Unexpected server response at END-OF-DATA: ' + resp)
                rcpt = match.group(1)
                try:
                    self._recipients.remove(rcpt)
                except ValueError:
                    raise DspamClientError(
                        'Message was accepted for unknown recipient ' + rcpt)
                self.results[rcpt] = {'accepted': True}
                logger.debug(
                    'Message accepted for recipient {} in LMTP mode'.format(
                        rcpt))
                if not len(self._recipients):
                    finished = True

        elif peek.startswith('X-DSPAM-Result:'):
            # Response is in summary format
            regex = re.compile('X-DSPAM-Result: ([^;]+); result="(\w+)"; '
                               'class="(\w+)"; probability=([\d\.]+); '
                               'confidence=([\d\.]+); signature=([\w,/]+)')
            finished = False
            while not finished:
                resp = self._read()
                match = regex.match(resp)
                if not match:
                    raise DspamClientError(
                        'Unexpected server response at END-OF-DATA: ' + resp)
                rcpt = match.group(1)
                try:
                    self._recipients.remove(rcpt)
                except ValueError:
                    raise DspamClientError(
                        'Message was accepted for unknown '
                        'recipient {}'.format(rcpt))

                # map results to their DSPAM classification result names
                fields = ('user', 'result', 'class',
                          'probability', 'confidence', 'signature')
                self.results[rcpt] = dict(zip(fields, match.groups()))
                if self.results[rcpt]['signature'] == 'N/A':
                    del(self.results[rcpt]['signature'])

                logger.debug(
                    'Message handled for recipient {} in DLMTP summary mode, '
                    'result is {}'.format(rcpt, match.group(2)))
                if not len(self._recipients):
                    # we received responses for all accepted recipients
                    finished = True
            # read final dot
            resp = self._read()
            if resp != '.':
                raise DspamClientError(
                    'Unexpected server response at END-OF-DATA: ' + resp)

        elif peek.startswith('X-Daemon-Classification:'):
            # Response is in stdout format
            finished = False
            message = ''
            while not finished:
                resp = self._read()
                if resp.startswith('X-Daemon-Classification:'):
                    if message != '':
                        # A new message body starts, store the previous one
                        rcpt = self._recipients.pop(0)
                        self.results[rcpt] = {
                            'result': result,
                            'message': message
                        }
                        logger.debug(
                            'Message handled for recipient {} in DLMTP '
                            'stdout mode, result is {}, message body '
                            'is {} chars'.format(rcpt, result, len(message)))
                        message = ''
                    # Remember next result
                    result = resp[25:]

                elif resp == '.':
                    # A single dot can signal end-of-data, or might be just
                    #   regular mail data.
                    self._socket.setblocking(False)
                    try:
                        # If _peek() succeeds, we did not reach end-of-data yet
                        #   so it was message content.
                        peek = self._peek(1)
                        message = message + '\r\n' + resp
                    except socket.error:
                        # reached end-of-data, store message and finish
                        finished = True
                        rcpt = self._recipients.pop(0)
                        # strip final newline
                        message = message[0:-2]
                        self.results[rcpt] = {
                            'result': result,
                            'message': message
                        }
                        logger.debug(
                            'Message accepted for recipient {} in DLMTP '
                            'stdout mode, result is {}, message body '
                            'is {} chars'.format(rcpt, result, len(message)))

                    self._socket.setblocking(True)

                else:
                    # regular message data
                    if message == '':
                        message = resp
                    else:
                        message = message + '\r\n' + resp

        else:
            raise DspamClientError(
                'Unexpected server response at END-OF-DATA: ' + resp)

    def rset(self):
        """
        Send LMTP RSET command and process the server response.

        """
        self._send('RSET\r\n')
        resp = self._read()
        if not resp.startswith('250'):
            logger.warn('Unexpected server response at RSET: ' + resp)
        self._recipients = []
        self.results = {}

    def quit(self):
        """
        Send LMTP QUIT command, read the server response and disconnect.

        """
        self._send('QUIT\r\n')
        resp = self._read()
        if not resp.startswith('221'):
            logger.warning('Unexpected server response at QUIT: ' + resp)
        self._socket.close()
        self._socket = None
        self._recipients = []
        self.results = {}

    def process(self, message, user):
        """
        Process a message.
        """
        if not self._socket:
            self.connect()
            self.lhlo()
        else:
            self.rset()

        if not self.dlmtp:
            raise DspamClientError('DLMTP mode not available')

        self.mailfrom(client_args='--process --deliver=summary')
        self.rcptto((user,))
        self.data(message)

        # check for valid result format
        if 'class' not in self.results[user]:
            raise DspamClientError(
                'Unexpected response format from server at END-OF-DATA, '
                'an error occured')

        return self.results[user]

    def classify(self, message, user):
        """
        Classify a message.

        """
        if not self._socket:
            self.connect()
            self.lhlo()
        else:
            self.rset()

        if not self.dlmtp:
            raise DspamClientError('DLMTP mode not available')

        self.mailfrom(client_args='--classify --deliver=summary')
        self.rcptto((user,))
        self.data(message)

        # check for valid result format
        if 'class' not in self.results[user]:
            raise DspamClientError(
                'Unexpected response format from server at END-OF-DATA, '
                'an error occured')

        return self.results[user]

    def train(self, message, user, class_):
        """
        Train DSPAM with a message.

        """
        raise NotImplementedError

    def retrain_message(self, message, class_, source='error'):
        """
        Correct an invalid classification.

        """
        raise NotImplementedError

    def retrain_signature(self, signature, class_, source='error'):
        """
        Correct an invalid classification.

        """
        raise NotImplementedError


if __name__ == '__main__':

    message = """Subject: Test mail
Message-ID: <TEST.1010101@example.org>
Date: Wed, 23 Jul 2003 23:30:00 +0200
From: Sender <sender@example.org>
To: Recipient <recipient@example.net>
Precedence: junk
MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit

This is a test mail.
Dot-stuffing test on next line:
.
The line above, you see?

"""

    logging.basicConfig(level=logging.DEBUG)

    # Config method 1
    c = DspamClient('inet:2424@localhost', 'test', 'leey2Pah')

    # Config method 2
    #c = DspamClient()
    #c.socket = 'inet:2424@localhost'
    #c.dlmtp_ident = 'test'
    #c.dlmtp_pass = 'leey2Pah'

    # Config method 3
    #DspamClient.socket = 'inet:2424@localhost'
    #DspamClient.dlmtp_ident = 'test'
    #DspamClient.dlmtp_pass = 'leey2Pah'
    #c = DspamClient()

    results = c.process(message, 'recipient@example.net')
    print('Classification results: ' + str(results))



# Copyright (c) 2012, Tom Hendrikx
# All rights reserved.
#
# See LICENSE for the license.

import ConfigParser
import datetime
import logging
import os.path
import re
import sys
import time

import Milter

from dspam import VERSION, utils
from dspam.client import *

logger = logging.getLogger(__name__)


class DspamMilter(Milter.Base):
    """
    A milter interface to the DSPAM daemon.

    This milter can be added to an MTA setup so messages can be inspected
    by a DSPAM server, and optionally rejected or quarantined based on the
    classification results.

    """

    # Constants defining possible return codes for compute_verdict()
    VERDICT_ACCEPT = 1
    VERDICT_QUARANTINE = 2
    VERDICT_REJECT = 3

    # Default configuration
    static_user = None
    headers = {
        'Processed': 0,
        'Confidence': 0,
        'Probability': 0,
        'Result': 0,
        'Signature': 0
    }
    header_prefix = 'X-DSPAM-'
    reject_classes = {'Blacklisted': 0, 'Blocklisted': 0, 'Spam': 0.9}
    quarantine_classes = {'Virus': 0}
    accept_classes = {'Innocent': 0, 'Whitelisted': 0}
    recipient_delimiter = '+'

    def __init__(self):
        """
        Create a new milter instance.

        """
        self.id = Milter.uniqueID()
        self.message = ''
        self.recipients = []
        self.dspam = None
        self.remove_headers = []
        if self.recipient_delimiter:
            self.recipient_delimiter_re = re.compile('[{}][^@]*'.format(
                re.escape(self.recipient_delimiter)))
        else:
            self.recipient_delimiter_re = None

    def connect(self, hostname, family, hostaddr):
        """
        Log new connections.

        """
        self.client_ip = hostaddr[0]
        self.client_port = hostaddr[1]
        self.time_start = time.time()
        logger.debug('<{}> Connect from {}[{}]:{}'.format(
            self.id, hostname, self.client_ip, self.client_port))
        return Milter.CONTINUE

    def envrcpt(self, rcpt, *params):
        """
        Send all recipients to DSPAM.

        """
        if rcpt.startswith('<'):
            rcpt = rcpt[1:]
        if rcpt.endswith('>'):
            rcpt = rcpt[:-1]
        if self.recipient_delimiter_re:
            rcpt = self.recipient_delimiter_re.sub('', rcpt)
        if rcpt not in self.recipients:
            self.recipients.append(rcpt)
            logger.debug('<{}> Received RCPT {}'.format(self.id, rcpt))
        return Milter.CONTINUE

    @Milter.noreply
    def header(self, name, value):
        """
        Store all message headers, optionally clean them up.

        This simply stores all message headers so we can send them to DSPAM.
        Additionally, headers that have the same prefix as the ones we're
        about to add are deleted.

        """
        self.message += "{}: {}\r\n".format(name, value)
        logger.debug('<{}> Received {} header'.format(self.id, name))
        if name.lower().startswith(self.header_prefix.lower()):
            self.remove_headers.append(name)
            logger.debug('<{}> Going to remove {} header'.format(
                self.id, name))
        return Milter.CONTINUE

    @Milter.noreply
    def eoh(self):
        """
        Store end of message headers.

        """
        self.message += "\r\n"
        return Milter.CONTINUE

    @Milter.noreply
    def body(self, block):
        """
        Store message body.

        """
        self.message += block
        logger.debug('<{}> Received {} bytes of message body'.format(
            self.id, len(block)))
        return Milter.CONTINUE

    def eom(self):
        """
        Send the message to DSPAM for classification and a return a milter
        response based on the results.

        If <DspamMilter>.static_user is set, that single DSPAM user account
        will be used for processing the message. If it is unset, all envelope
        recipients will be passed to DSPAM, and the final decision is based on
        the least invasive result in all their classification results.

        """
        for header in self.remove_headers:
            self.chgheader(header, 1, '')
            logger.info('<{}> Removing existing {} header'.format(
                self.id, header))

        queue_id = self.getsymval('i')
        logger.debug(
            '<{}> Sending message with MTA queue id {} to DSPAM'.format(
                self.id, queue_id))

        try:
            if not self.dspam:
                self.dspam = DspamClient()
                self.dspam.connect()
                self.dspam.lhlo()
                if not self.dspam.dlmtp:
                    logger.warning(
                        '<{}> Connection to DSPAM is established, but DLMTP '
                        'seems unavailable'.format(self.id))
            else:
                self.dspam.rset()
        except DspamClientError, err:
            logger.error(
                '<{}> An error ocurred while connecting to DSPAM: {}'.format(
                    self.id, err))
            return Milter.TEMPFAIL

        try:
            self.dspam.mailfrom(client_args='--process --deliver=summary')
            if self.static_user:
                self.dspam.rcptto((self.static_user,))
            else:
                self.dspam.rcptto(self.recipients)
            self.dspam.data(self.message)
        except DspamClientError, err:
            logger.error(
                '<{}> An error ocurred while talking to DSPAM: {}'.format(
                    self.id, err))
            return Milter.TEMPFAIL

        # Clear caches
        self.message = ''
        self.recipients = []

        # With multiple recipients, if different verdicts were returned, always
        #   use the 'lowest' verdict as final, so mail is not lost unexpected.
        final_verdict = None
        for rcpt in self.dspam.results:
            results = self.dspam.results[rcpt]
            logger.info('<{}> DSPAM returned results for RCPT {}: {}'.format(
                self.id, rcpt,
                ' '.join('{}={}'.format(k, v) for
                         k, v in results.iteritems())))
            verdict = self.compute_verdict(results)
            if final_verdict is None or verdict < final_verdict:
                final_verdict = verdict
                final_results = results

        if final_verdict == self.VERDICT_REJECT:
            logger.info(
                '<{0}> Rejecting message with queue id {1} based on DSPAM '
                'results: user={2[user]} class={2[class]} '
                'confidence={2[confidence]}'.format(
                    self.id, queue_id, final_results))
            self.setreply('550', '5.7.1', 'Message is {0[class]}'.format(
                final_results))
            return Milter.REJECT
        elif final_verdict == self.VERDICT_QUARANTINE:
            logger.info(
                '<{0}> Quarantining message with queue id {1} based on DSPAM '
                'results: user={2[user]} class={2[class]} '
                'confidence={2[confidence]}'.format(
                    self.id, queue_id, final_results))
            self.add_dspam_headers(final_results)
            self.quarantine('Message is {0[class]} according to DSPAM'.format(
                final_results))
            return Milter.ACCEPT
        else:
            logger.info(
                '<{0}> Accepting message with queue id {1} based on DSPAM '
                'results: user={2[user]} class={2[class]} '
                'confidence={2[confidence]}'.format(
                    self.id, queue_id, final_results))
            self.add_dspam_headers(final_results)
            return Milter.ACCEPT

    def close(self):
        """
        Log disconnects.

        """
        time_spent = time.time() - self.time_start
        logger.debug(
            '<{}> Disconnect from [{}]:{}, time spent {:.3f} seconds'.format(
                self.id, self.client_ip, self.client_port, time_spent))
        return Milter.CONTINUE

    def compute_verdict(self, results):
        """
        Match results to the configured reject, quarantine and accept classes,
        and return a verdict based on that.

        The verdict classes are matched in the order: reject_classes,
        quarantine_classes, accept_classes. This means that you can configure
        different verdicts for different confidence results, for instance:
        reject_classes= Spam:0.99       # Reject obvious spam
        quarantine_classes = Spam:0.7   # Quarantine spam with confidence
                                        #   between 0.7 and 0.99
        accept_classes = Spam           # Accept low confidence spam (good
                                        #   for FP and retraining)

        Args:
        results -- A results dictionary from DspamClient.

        """
        if results['class'] in self.reject_classes:
            threshold = self.reject_classes[results['class']]
            if float(results['confidence']) >= threshold:
                logger.debug(
                    '<{0}> Suggesting to reject the message based on DSPAM '
                    'results: user={1[user]}, class={1[class]}, '
                    'confidence={1[confidence]}'.format(self.id, results))
                return self.VERDICT_REJECT

        if results['class'] in self.quarantine_classes:
            threshold = self.quarantine_classes[results['class']]
            if float(results['confidence']) >= threshold:
                logger.debug(
                    '<{0}> Suggesting to quarantine the message based on '
                    'DSPAM results: user={1[user]}, class={1[class]}, '
                    'confidence={1[confidence]}'.format(self.id, results))
                return self.VERDICT_QUARANTINE

        if results['class'] in self.accept_classes:
            threshold = self.accept_classes[results['class']]
            if float(results['confidence']) >= threshold:
                logger.debug(
                    '<{0}> Suggesting to accept the message based on DSPAM '
                    'results: user={1[user]}, class={1[class]}, '
                    'confidence={1[confidence]}'.format(self.id, results))
                return self.VERDICT_ACCEPT

        logger.debug(
            '<{0}> Suggesting to accept the message, no verdict class matched '
            'DSPAM results: user={1[user]}, class={1[class]}, '
            'confidence={1[confidence]}'.format(self.id, results))
        return self.VERDICT_ACCEPT

    def add_dspam_headers(self, results):
        """
        Format DSPAM headers with passed results, and add them to the message.

        Args:
        results -- A results dictionary from DspamClient.
        """
        for header in self.headers:
            hname = self.header_prefix + header
            if header.lower() in results:
                hvalue = results[header.lower()]
                logger.debug(
                    '<{}> Adding header {}: {}'.format(self.id, hname, hvalue))
                self.addheader(hname, hvalue)
            elif header == 'Processed':
                # X-DSPAM-Processed: Wed Dec 12 02:19:23 2012
                hvalue = datetime.datetime.now().strftime(
                    '%a %b %d %H:%M:%S %Y')
                logger.debug(
                    '<{}> Adding header {}: {}'.format(self.id, hname, hvalue))
                self.addheader(hname, hvalue)
            else:
                logger.warning(
                    '<{}> Not adding header {}, no data available in '
                    'DSPAM results'.format(self.id, hname))


class DspamMilterDaemon(object):
    """
    Run the Milter as a UNIX daemon process.

    """

    # Default configuration
    socket = 'inet:2425@localhost'
    timeout = 300
    loglevel = 'INFO'
    #pidfile = '/var/run/dspam/dspam-milter.pid'
    pidfile = '/tmp/dspam-milter.pid'
    daemonize = True

    def run(self, config_file=None):
        utils.log_to_syslog()
        logger.info('DSPAM Milter startup (v{})'.format(VERSION))
        if config_file is not None:
            self.configure(config_file)
        if self.daemonize:
            utils.daemonize(self.pidfile)
        Milter.factory = DspamMilter
        Milter.runmilter('DspamMilter', self.socket, self.timeout)
        logger.info('DSPAM Milter shutdown (v{})'.format(VERSION))
        logging.shutdown()

    def configure(self, config_file):
        """
        Parse configuration, and setup objects to use it.

        """
        cfg = ConfigParser.RawConfigParser()
        try:
            cfg.readfp(open(config_file))
        except IOError, err:
            logger.critical(
                'Error while reading config file {}: {}'.format(
                    config_file, err.strerror))
            sys.exit(1)
        logger.info('Parsed config file ' + config_file)

        # Extract user-defined log level from configuration
        if cfg.has_option('milter', 'loglevel'):
            loglevel = cfg.get('milter', 'loglevel')
            loglevel_numeric = getattr(logging, loglevel.upper(), None)
            if not isinstance(loglevel_numeric, int):
                logger.critical(
                    'Config contains unsupported loglevel: ' + loglevel)
                exit(1)
            rl = logging.getLogger()
            rl.setLevel(loglevel_numeric)
            logger.debug(
                'Config option applied: milter->loglevel: {}'.format(loglevel))

        # Apply all config options to their respective classes
        section_class_map = {
            'milter': self,
            'dspam': DspamClient,
            'classification': DspamMilter,
        }
        for section in cfg.sections():
            try:
                class_ = section_class_map[section]
            except KeyError:
                logger.warning('Config contains unknown section: ' + section)
                continue
            logger.debug('Handling config section: ' + section)

            dict_options = [
                'headers',
                'reject_classes',
                'quarantine_classes',
                'accept_classes'
            ]
            for option in cfg.options(section):
                # Kludge: static_user needs to be set on the milter,
                #   not on the client
                if section == 'dspam' and option == 'static_user':
                    value = cfg.get('dspam', 'static_user')
                    DspamMilter.static_user = value
                    logger.debug(
                        'Config option applied: dspam->static_user: {}'.format(
                            value))
                    continue

                if not hasattr(class_, option):
                    logger.warning(
                        'Config contains unknown option: {}:{}'.format(
                            section, option))
                    continue

                value = cfg.get(section, option)
                if option in dict_options:
                    value = utils.config_str2dict(value)
                elif value.lower() in ['false', 'no']:
                    value = False
                elif value.lower() in ['true', 'yes']:
                    value = True

                setattr(class_, option, value)
                logger.debug(
                    'Config option applied: {}->{}: {}'.format(
                        section, option, value))
        logger.debug('Configuration completed')


def main(config_file=None):
    d = DspamMilterDaemon()
    d.run(config_file)

if __name__ == "__main__":
    main()
